#define EBATWO
#region using
using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
using System.Reflection;
using MVR.FileManagementSecure;
#endregion

namespace Cotyounoyume{
	public class EBATCore : MVRScript
	{
		public float version = 1.07f;

		#region common parameters
		public Atom atom;
		public SimpleJSON.JSONClass settings = new SimpleJSON.JSONClass();
		public string myPath;
		public bool maintenanceAlert = false;
		public float lastTouchingTime = 0f;
		
		protected float openingTime;
		protected float currentTime;
		public int gauge = 0;
		public int curStage = -1;
		public int preStage = -1;
		protected JSONStorableBool atOnce;
		protected string defaultSettingsPath = "Custom/Scripts/cotyounoyume/ExpressionBlushingAndTears/Config/defaultSettings.json";
		protected string currentSettingsPath = "Custom/Scripts/ExpressionBlushingAndTears_currentSettings.json";
		protected string myClassName = "EBATCore";

		public string pluginLabel;
		public bool disableLog = true;
		public int[] stageThreshold;
		public bool isBreathActive = false;
		public bool isDuringOrgasm = false;
		public bool isDuringAfterglow = false;
		public string myName = "";
		
		public Expression expression;
		public Blushing blushing;
		public Tears tears;
		public EyesControl eyesControl;
		public Belly belly;
		public Etc etc;
		public BodyMorph bodyMorph;
#if	(EBATWO)
		public Orgasm orgasm;
#endif
		public Sound sound;
		#endregion
		
		public override void Init(){
			try{
				if(atOnce == null){
					logger("Plugin installed");
					atOnce = new JSONStorableBool("atOnce", false);
				}else{
					logger("already loaded.");
				}
                if (containingAtom.type != "Person")
                {
                    SuperController.LogError("This plug-in have to be added to person type atom.");
                    return;
                }
				atom = containingAtom;
				myName = atom.name;
				logger("atom name: " + myName);
				myPath = GetPluginPath();
				logger("myPath: " + myPath);
				
				expression  = new Expression(this);
				blushing    = new Blushing(this);
				tears       = new Tears(this);
				bodyMorph   = new BodyMorph(this);
				eyesControl = new EyesControl(this);
				belly       = new Belly(this);
				etc         = new Etc(this);
				sound       = new Sound(this);
#if	(EBATWO)				
				orgasm      = new Orgasm(this);
#endif
			}catch (Exception e){
				loggerE(e);
			}
		}

		public void syncParameter(){
			isDuringOrgasm = false;
			isDuringAfterglow = false;
#if	(EBATWO)
			isDuringOrgasm = orgasm.isDuringOrgasm;
			isDuringAfterglow = orgasm.isDuringAfterglow;
#endif
		}
		
		public void Start(){
			try{
				#region at onece: base
				if(configFileExist() == false){
					loadSettings("default");
					saveConfig();
				}
				loadSettings("current");
				applySettings();
//				removeConfigMenu();
//				removeConfigMenuForce();
				setupUI();
				
				openingTime = Time.time;
				logger("openingTime = " + openingTime);
				logger("myPath:" + myPath);
				gauge = 0;
				isBreathActive = false;
				#endregion

				maintenanceSequence();
				belly.storePositionAndRotation();
				tears.unequipTearsAll();
				bodyMorph.stopBodyMorph();
				expression.stopExpression();
				
//				surveyStorables(1);
			}catch (Exception e){
				loggerE(e);
			}			
		}

		public bool configFileExist(){
			try{
				if(FileManagerSecure.FileExists(currentSettingsPath)){
					return true;
				}else{
					return false;
				}
			}catch (Exception e){
				loggerE(e);
				return false;
			}
		}
		
		public void loadSettings(string type){
			try{
				defaultSettingsPath = myPath + "/" + "Config/defaultSettings.json";
				string path = (type == "default") ? defaultSettingsPath : currentSettingsPath;
				settings = SuperController.singleton.LoadJSON(path) as JSONClass;
				if(version != settings["General"]["Version"].AsFloat){
					defaultSettingsPath = myPath + "/" + "Config/defaultSettings.json";
					settings = SuperController.singleton.LoadJSON(defaultSettingsPath) as JSONClass;					
				}
				
				#region general settings from json
				pluginLabel = settings["General"]["Label"];
				disableLog = settings["General"]["DebugMessageOff"].AsBool;
				expression.disableLog = disableLog;
				blushing.disableLog = disableLog;
				tears.disableLog = disableLog;
				bodyMorph.disableLog = disableLog;
				eyesControl.disableLog = disableLog;
				belly.disableLog = disableLog;
				etc.disableLog = disableLog;
				sound.disableLog = disableLog;
				
				stageThreshold = new int[6];
				stageThreshold[0] = 0;
				for(int i = 0; i < 6; i++){
					stageThreshold[i] = settings["StageThreshold"][("Threshold" + i)]["Value"].AsInt;				
				}
				#endregion
				
				#region 
				expression.loadSettings();
				blushing.loadSettings();
				tears.loadSettings();
				bodyMorph.loadSettings();
				eyesControl.loadSettings();
				belly.loadSettings();
				etc.loadSettings();
				sound.loadSettings();
				
				loadSettingsAlt();
				
				#endregion
			}catch (Exception e){
				loggerE(e);
			}
		}

		public void loadSettingsAlt(){
#if (EBATWO)
			orgasm.loadSettings();
			orgasm.applySettings();
			orgasm.maintenance();
#endif
		}
		
		public void applySettings(){
			try{
				pluginLabelJSON.val = settings["General"]["Label"];

				#region
				expression.applySettings();
				blushing.applySettings();
				tears.applySettings();
				bodyMorph.applySettings();
				eyesControl.applySettings();
				belly.applySettings();
				etc.applySettings();
				#endregion
			}catch (Exception e){
				loggerE(e);
			}
		}
		
		public void maintenanceSequence(){
			try{
				#region initialize trigger
				resetTriggers();
				RegisterActions();
				#endregion
				
				#region maintenance
				expression.maintenance();
				blushing.maintenance();
				tears.maintenance();
				bodyMorph.maintenance();
				eyesControl.maintenance();
				belly.maintenance();
				etc.maintenance();
				#endregion
			}catch (Exception e){
				loggerE(e);
			}
		}
		
		public void saveConfig(){
			try{
				SuperController.singleton.SaveJSON(settings, currentSettingsPath);
			}catch (Exception e){
				loggerE(e);
			}
		}
		
		#region trigger functions
		public void RegisterActions(){
			try{
//				RegisterActionPair("LipTrigger", "OnLipTouch", "Lip");				
				RegisterActionPair("MouthTrigger", "OnMouthTouch", "Mouth");				
				RegisterActionPair("ThroatTrigger", "OnThroatTouch", "Throat");				
				RegisterActionPair("lNippleTrigger", "OnNippleTouch", "Nipple");
				RegisterActionPair("rNippleTrigger", "OnNippleTouch", "Nipple");
				RegisterActionPair("LabiaTrigger", "OnLabiaTouch", "Labia");
				RegisterActionPair("VaginaTrigger", "OnVaginaTouch", "Vagina");
				RegisterActionPair("DeepVaginaTrigger", "OnVaginaDeepTouch", "VaginaDeep");
				RegisterActionPair("DeeperVaginaTrigger", "OnVaginaDeeperTouch", "VaginaDeeper");
			}catch (Exception e){
				loggerE(e);
			}
		}
		
		public void RegisterActionPair(string collisionTrigger, string action, string position){
			try{
				JSONStorableAction tmpAction = new JSONStorableAction(action, () =>{
					OnErogenousZoneTouch(position);
				});
				RegisterAction(tmpAction);
				createTrigger(collisionTrigger, action);
			}catch (Exception e){
				loggerE(e);
			}
		}
		
		public void resetTriggers() {
			try{
				var table = new Dictionary<string, int>{
//					{"OnLipTouch", 1},
					{"OnMouthTouch", 1},
					{"OnThroatTouch", 1},
					{"OnNippleTouch", 1},
					{"OnLabiaTouch", 1},
					{"OnVaginaTouch", 1},
					{"OnVaginaDeepTouch", 1},
					{"OnVaginaDeeperTouch", 1}
				};
//				string[] cTriggerNames = {"LipTrigger", "MouthTrigger", "ThroatTrigger","rNippleTrigger", "lNippleTrigger", "LabiaTrigger", "VaginaTrigger", "DeepVaginaTrigger", "DeeperVaginaTrigger"};
				string[] cTriggerNames = {"MouthTrigger", "ThroatTrigger","rNippleTrigger", "lNippleTrigger", "LabiaTrigger", "VaginaTrigger", "DeepVaginaTrigger", "DeeperVaginaTrigger"};
				foreach(string cTriggerName in cTriggerNames){
					logger("cTriggerName = " + cTriggerName);
					CollisionTrigger cTrigger = atom.GetStorableByID(cTriggerName) as CollisionTrigger;
					JSONClass triggerClass = cTrigger.trigger.GetJSON();
					JSONArray triggerArray = triggerClass["startActions"].AsArray;
					for (int i=0;i< triggerArray.Count;i++) {
						logger("triggerArray[" + i + "][name]=" + triggerArray[i]["name"].Value);
						if (table.ContainsKey(triggerArray[i]["name"].Value)){
							triggerArray.Remove(i);
						}
					}
					cTrigger.trigger.RestoreFromJSON(triggerClass);
				}
			}catch (Exception e){
				loggerE(e);
			}
		}
		
		public void createTrigger(string triggerCollisionTriggerName, string scriptTriggerName){
			try{
				CollisionTrigger cTrigger = atom.GetStorableByID(triggerCollisionTriggerName) as CollisionTrigger;
				cTrigger.enabled = true;
				TriggerActionDiscrete startTrigger;
				startTrigger=cTrigger.trigger.CreateDiscreteActionStartInternal();
				startTrigger.name = scriptTriggerName;
				startTrigger.receiverAtom = atom;
				startTrigger.receiver = GetPluginStorableById(atom, myClassName);
				startTrigger.receiverTargetName = scriptTriggerName;
			}catch (Exception e){
				loggerE(e);
			}
		}
	
		// Function taken from VAMDeluxe's code
		static JSONStorable GetPluginStorableById(Atom atom, string id){
			try{
				string storableIdName = atom.GetStorableIDs().FirstOrDefault((string storeId) =>{
					if (string.IsNullOrEmpty(storeId)){
						return false;
					}
					return storeId.Contains(id);
				});
				if (storableIdName == null){
					return null;
				}
				return atom.GetStorableByID(storableIdName);
			}catch (Exception e){
//				loggerE(e);
				return null;
			}
		}
		#endregion		
		
		#region engine
		public void OnErogenousZoneTouch(string zone){ 
			try{
				logger("Touch: " + zone);
				currentTime = Time.time;
				lastTouchingTime = currentTime;
				sound.playSound("moan", Time.time);
				updateAll();			

				gauge++;
				isStageChanged();
				
				if(isBreathActive == false){
					isBreathActive = true;
				}
				
				orgasmFunctions("orgasmCheck", currentTime);
			}catch (Exception e){
				loggerE(e);
			}
		}
		
		public void OnStageChange(){
			try{
				tears.cm.unequipClothes(preStage);
				tears.cm.equipClothes(curStage);
			}catch (Exception e){
				loggerE(e);
			}			
		}
		
		public void isStageChanged(){
			try{
				if(curStage != getStage()){
					preStage = curStage;
					curStage = getStage();
					logger("stage changed: " + preStage + " => " + curStage);
					OnStageChange();
				}
			}catch(Exception e){
				loggerE(e);
			}
		}
		
		public void orgasmFunctions(string func, float time){
#if (EBATWO)
			if(func == "orgasmCheck"){
				orgasm.orgasmCheck(time);
			}else if(func == "update"){
				orgasm.update(time);
			}else if(func == "orgasmControl"){
				orgasm.orgasmControl(time);
			}else if(func == "orgasmStart"){
				orgasm.orgasmStart();
			}else if(func == "orgasmFinish"){
				orgasm.orgasmFinish();
			}
#endif
		}
		
		public bool isOrgasmFunctionEnabled(){
			bool flag = false;
#if (EBATWO)
			flag = true;
#endif
			return flag;
		}
		
		public void updateAll(){
			try{
				expression.update(currentTime);
				blushing.update(currentTime);
				tears.update(currentTime);
				bodyMorph.update(currentTime);
				eyesControl.update(currentTime);
				orgasmFunctions("update", currentTime);
			}catch (Exception e){
				loggerE(e);
			}
		}
				
        public void Update(){
            try{
				if(Time.time > currentTime + 10){
					currentTime = Time.time;
					logger("start plug-in or load scene.");
					maintenanceSequence();
				}
				syncParameter();
				string currentName = atom.name;
				if(myName != currentName){
					logger("old name:" + myName + ", new name:" + currentName);
					myName = currentName;
					maintenanceAlert = true;
				}
				
				if(maintenanceAlert){
					maintenanceAlert = false;
					Init();
					removeConfigMenu();
					Start();
					return;
				}
				currentTime = Time.time;

				orgasmFunctions("orgasmControl", currentTime);
				blushing.manage(currentTime);
				tears.manage(currentTime);
				bodyMorph.manage(currentTime);
				expression.manage(currentTime);
				eyesControl.manage(currentTime);
				belly.manage(currentTime);
				
				if(expression.isExpressionActive == false && blushing.isBlushingActive == false && tears.isTearsActive == false){
					if(isDuringOrgasm == false){
						gauge = 0;
						isStageChanged();
					}
				}
				
				if(isOrgasmFunctionEnabled() == false &&
					lastTouchingTime + longestDurationTime() < currentTime){
					gauge = 0;
					isStageChanged();
				}
			}catch (Exception e){
				loggerE(e);
			}
		}

		public float longestDurationTime(){
			float eTime = expression.expressionClosingTimeOffset1 + expression.expressionClosingTimeOffset2;
			float bTime = blushing.blushClosingTimeOffset;
			float tTime = tears.tearsClosingTimeOffset;
			float tmpTime = (eTime > bTime) ? eTime : bTime;
			float result = (tmpTime > tTime) ? tmpTime : tTime;
			return result;
		}
		
		public int getStage(){
			try{
				int stage = 0;
				if(gauge < stageThreshold[1]){
					stage = 0;
				}else if(gauge < stageThreshold[2]){
					stage = 1;
				}else if(gauge < stageThreshold[3]){
					stage = 2;
				}else if(gauge < stageThreshold[4]){
					stage = 3;
				}else if(gauge < stageThreshold[5]){
					stage = 4;
//				}else if(isDuringAfterglow == false){
//					stage = 5;
				}else{
					stage = 5;
//					stage = 6;
				}
				return stage;
			}catch (Exception e){
				loggerE(e);
				return -1;
			}
		}
		
		#endregion
		
		#region etc
		public void logger(string str){
			if(disableLog == false){
				SuperController.LogMessage(str);
			}
		}
		
		public void loggerE(Exception e){
			SuperController.LogMessage("*** Exception:" + e);
		}
		
		public void debug(string str){
			SuperController.LogMessage("debug: " + str);
		}
		
		// Function taken from VAMDeluxe's code
		private string GetPluginPath(){
			string id = name.Substring(0, name.IndexOf('_'));
            string filename = manager.GetJSON()["plugins"][id].Value;
            return filename.Substring(0, filename.LastIndexOfAny(new char[] { '/', '\\' }));
		}
		#endregion
		
		public void OnDestroy(){
			expression.stopExpression();
			bodyMorph.stopBodyMorph();
		}
		
		public void surveyStorables(int lvl){
			List<string> ids = new List<string>();
			ids = atom.GetStorableIDs();
			for(int i = 0; i < ids.Count; i++){
				SuperController.LogMessage("id: " + ids[i]);
				List<string> names = new List<string>();
				
				if(lvl > 0){
					names = atom.GetStorableByID(ids[i]).GetBoolParamNames();
					for(int j = 0; j < names.Count; j++){
						SuperController.LogMessage("  bool: " + names[j]);
					}
				}
					
				if(lvl > 1){
					names = atom.GetStorableByID(ids[i]).GetAllParamAndActionNames();
					for(int j = 0; j < names.Count; j++){
						SuperController.LogMessage("  ParamAndAction: " + names[j]);
					}					
				}

				if(lvl > 1){
					names = atom.GetStorableByID(ids[i]).GetAllFloatAndColorParamNames();
					for(int j = 0; j < names.Count; j++){
						SuperController.LogMessage("  FloatAndColorParam: " + names[j]);
					}
				}
			}
		}
		
		public void setupUI(){
			setupUIActionButtons();
			setupUIExpandToggleMenu();
			if(expandToggleMenu){
				setupUIToggleButtons();
			}
			setupUIExpandSlider1Menu();
			if(expandSlider1Menu){
				setupUISliders1();
			}
			setupUIExpandSoundMenu();
			if(expandSoundMenu){
				setupUISound();
			}
			setupUIExpandMorphEMenu();
			if(expandMorphEMenu){
				setupUIMorphsE();
			}
			setupUIExpandMorphOMenu();
			if(expandMorphOMenu){
				setupUIMorphsO();
			}
			
			setupUIExpandSlider2Menu();
			if(expandSlider2Menu){
				setupUISliders2();
			}
#if (EBATWO)
			setupUIExpandSlider3Menu();
			if(expandSlider3Menu){
				setupUISliders3();
			}
#endif
		}
		
		public void setupUIActionButtons(){
			try{
				resetButton = CreateButton("Load Default", false);
				resetButton.button.onClick.AddListener(() =>{
					expression.stopExpression();
					bodyMorph.stopBodyMorph();
					logger("load default settings.");
					loadSettings("default");
					saveConfig();
					removeConfigMenu();
					Start();
				});	
				resetButton.buttonColor = Color.cyan;
				
				 toggleDebug = new JSONStorableBool("Debug Message Off", disableLog, (bool toggle) => {
					disableLog = disableLog ? false : true;
					expression.disableLog = disableLog;
					blushing.disableLog = disableLog;
					settings["General"]["DebugMessageOff"] = disableLog ? "true" : "false";
					if(disableLog == false){
						logger("disableLog: " + disableLog);
					}
					saveConfig();
				});
				CreateToggle((toggleDebug), true);
				
				copySoundButton = CreateButton("Copy Sound File", false);
				copySoundButton.button.onClick.AddListener(() =>{
					sound.copySound();
				});	
				copySoundButton.buttonColor = Color.cyan;
				
				spacer03 = CreateSpacer(true);
				spacer03.height = 50f;
				
#if (EBATWO)
				orgasmButton = CreateButton("Make her Orgasm", false);
				orgasmButton.button.onClick.AddListener(() =>{
					gauge = stageThreshold[5];
					OnErogenousZoneTouch("DeeperVagina");
					updateAll();
//					orgasm.orgasmStart();
					orgasmFunctions("orgasmStart", 0f);
				});	
				orgasmButton.buttonColor = Color.cyan;

				stopOrgasmButton = CreateButton("Stop her Orgasm", true);
				stopOrgasmButton.button.onClick.AddListener(() =>{
					gauge = stageThreshold[0];
//					orgasm.orgasmFinish();
					orgasmFunctions("orgasmFinish", 0f);
				});	
				stopOrgasmButton.buttonColor = Color.cyan;
#endif
				spacer01 = CreateSpacer(false);
				spacer01.height = 16f;
				spacer02 = CreateSpacer(true);
				spacer02.height = 16f;
			}catch (Exception e){
				loggerE(e);
			}
		}

		public void setupUIExpandToggleMenu(){
			try{
				if(expandToggleMenu == false){
					expandToggleButton = CreateButton("[+]Expand Toggles for Main Features", false);
					expandToggleButton.button.onClick.AddListener(() =>{
						removeConfigMenu();
						expandToggleMenu = true;
						setupUI();
					});	
					expandToggleButton.buttonColor = Color.white;
				}else{
					collapseToggleButton = CreateButton("[-]Collapse Toggles for Main Features", false);
					collapseToggleButton.button.onClick.AddListener(() =>{
						removeConfigMenu();
						expandToggleMenu = false;
						setupUI();
					});	
					collapseToggleButton.buttonColor = Color.white;					
				}

//				spacerA1 = CreateSpacer(false);
//				spacerA1.height = 4f;
				spacerA2 = CreateSpacer(true);
				spacerA2.height = 50f;

			}catch (Exception e){
				loggerE(e);
			}
		}
		
		public void setupUIExpandMorphOMenu(){
			try{
				if(expandMorphOMenu == false){
					expandMorphOButton = CreateButton("[+]Expand MorphForOrgasm Menu", false);
					expandMorphOButton.button.onClick.AddListener(() =>{
						removeConfigMenu();
						expandMorphOMenu = true;
						setupUI();
					});	
					expandMorphOButton.buttonColor = Color.white;					
				}else{
					collapseMorphOButton = CreateButton("[-]Collapse MorphForOrgasm Menu", false);
					collapseMorphOButton.button.onClick.AddListener(() =>{
						removeConfigMenu();
						expandMorphOMenu = false;
						expression.stopExpression();
						setupUI();
					});	
					collapseMorphOButton.buttonColor = Color.white;					
				}

				/*
				spacerB1 = CreateSpacer(false);
				spacerB1.height = 16f;
				*/
				spacerB2 = CreateSpacer(true);
				spacerB2.height = 50f;
			}catch (Exception e){
				loggerE(e);
			}
		}
		
		public void setupUIExpandMorphEMenu(){
			try{
				if(expandMorphEMenu == false){
					expandMorphEButton = CreateButton("[+]Expand MorphForEmotions Menu", false);
					expandMorphEButton.button.onClick.AddListener(() =>{
						removeConfigMenu();
						expandMorphEMenu = true;
						setupUI();
					});	
					expandMorphEButton.buttonColor = Color.white;					
				}else{
					collapseMorphEButton = CreateButton("[-]Collapse MorphForEmotions Menu", false);
					collapseMorphEButton.button.onClick.AddListener(() =>{
						removeConfigMenu();
						expandMorphEMenu = false;
						expression.stopExpression();
						setupUI();
					});	
					collapseMorphEButton.buttonColor = Color.white;					
				}
				
				/*
				spacerF1 = CreateSpacer(false);
				spacerF1.height = 16f;
				*/
				spacerF2 = CreateSpacer(true);
				spacerF2.height = 50f;
			}catch (Exception e){
				loggerE(e);
			}
		}
		
		public void setupUIExpandSlider1Menu(){
			try{
				if(expandSlider1Menu == false){
					expandSlider1Button = CreateButton("[+]Expand Sliders for Main Features", false);
					expandSlider1Button.button.onClick.AddListener(() =>{
						removeConfigMenu();
						expandSlider1Menu = true;
						setupUI();
					});	
					expandSlider1Button.buttonColor = Color.white;
				}else{
					collapseSlider1Button = CreateButton("[-]Collapse Sliders for Main Features", false);
					collapseSlider1Button.button.onClick.AddListener(() =>{
						removeConfigMenu();
						expandSlider1Menu = false;
						setupUI();
					});	
					collapseSlider1Button.buttonColor = Color.white;					
				}
				
				/*
				spacerC1 = CreateSpacer(false);
				spacerC1.height = 16f;
				*/
				spacerC2 = CreateSpacer(true);
				spacerC2.height = 50f;
			}catch (Exception e){
				loggerE(e);
			}
		}
		
		public void setupUIExpandSlider2Menu(){
			try{
				if(expandSlider2Menu == false){
					expandSlider2Button = CreateButton("[+]Expand Sliders for Stage Thresholds", false);
					expandSlider2Button.button.onClick.AddListener(() =>{
						removeConfigMenu();
						expandSlider2Menu = true;
						setupUI();
					});	
					expandSlider2Button.buttonColor = Color.white;					
				}else{
					collapseSlider2Button = CreateButton("[-]Collapse Sliders for Stage Thresholds", false);
					collapseSlider2Button.button.onClick.AddListener(() =>{
						removeConfigMenu();
						expandSlider2Menu = false;
						setupUI();
					});	
					collapseSlider2Button.buttonColor = Color.white;					
				}
				
				/*
				spacerD1 = CreateSpacer(false);
				spacerD1.height = 16f;
				*/
				spacerD2 = CreateSpacer(true);
				spacerD2.height = 50f;
			}catch (Exception e){
				loggerE(e);
			}
		}
		
		public void setupUIExpandSlider3Menu(){
			try{
				if(expandSlider3Menu == false){
					expandSlider3Button = CreateButton("[+]Expand Sliders for Orgasm Features", false);
					expandSlider3Button.button.onClick.AddListener(() =>{
						removeConfigMenu();
						expandSlider3Menu = true;
						setupUI();
					});	
					expandSlider3Button.buttonColor = Color.white;					
				}else{
					collapseSlider3Button = CreateButton("[-]Collapse Sliders for Orgasm Features", false);
					collapseSlider3Button.button.onClick.AddListener(() =>{
						removeConfigMenu();
						expandSlider3Menu = false;
						setupUI();
					});	
					collapseSlider3Button.buttonColor = Color.white;					
				}
				
				/*
				spacerE1 = CreateSpacer(false);
				spacerE1.height = 16f;
				*/
				spacerE2 = CreateSpacer(true);
				spacerE2.height = 50f;
			}catch (Exception e){
				loggerE(e);
			}
		}
			
		public void setupUIExpandSoundMenu(){
			try{
				if(expandSoundMenu == false){
					expandSoundButton = CreateButton("[+]Expand Sound Menu", false);
					expandSoundButton.button.onClick.AddListener(() =>{
						removeConfigMenu();
						expandSoundMenu = true;
						setupUI();
					});	
					expandSoundButton.buttonColor = Color.white;					
				}else{
					collapseSoundButton = CreateButton("[-]Collapse Sound Menu", false);
					collapseSoundButton.button.onClick.AddListener(() =>{
						removeConfigMenu();
						expandSoundMenu = false;
						setupUI();
					});	
					collapseSoundButton.buttonColor = Color.white;					
				}
				
				/*
				spacerE1 = CreateSpacer(false);
				spacerE1.height = 16f;
				*/
				spacerG2 = CreateSpacer(true);
				spacerG2.height = 50f;
			}catch (Exception e){
				loggerE(e);
			}
		}
			
		public void setupUIToggleButtons(){
			try{
		
				 toggleAE = new JSONStorableBool("Auto Expression Off", etc.autoExpressionOffGlobal, (bool toggle) => {
					etc.autoExpressionOffGlobal = etc.autoExpressionOffGlobal ? false : true;
					settings["General"]["AutoExpressionsOff"] = etc.autoExpressionOffGlobal ? "true" : "false";
					atom.GetStorableByID("AutoExpressions").SetBoolParamValue("enabled", !(etc.autoExpressionOffGlobal));
					saveConfig();
				});
				CreateToggle((toggleAE), false);
				
				 toggleAB = new JSONStorableBool("Auto Blink Off", etc.autoBlinkOffGlobal, (bool toggle) => {
					etc.autoBlinkOffGlobal = etc.autoBlinkOffGlobal ? false : true;
					settings["General"]["AutoBlinkOff"] = etc.autoBlinkOffGlobal ? "true" : "false";
					atom.GetStorableByID("EyelidControl").SetBoolParamValue("blinkEnabled", !(etc.autoBlinkOffGlobal));
					saveConfig();
				});
				CreateToggle((toggleAB), true);

				 toggleExpression = new JSONStorableBool("Expression On", expression.expressionOnGlobal, (bool toggle) => {
					expression.expressionOnGlobal = expression.expressionOnGlobal ? false : true;
					settings["General"]["ExpressionOn"] = expression.expressionOnGlobal ? "true" : "false";
					if(expression.expressionOnGlobal){
						expression.update(Time.time);
					}else{
						expression.stopExpression();
					}
					saveConfig();
				});
				CreateToggle((toggleExpression), false);
				
				 toggleSounds = new JSONStorableBool("Sound On", sound.soundOnGlobal, (bool toggle) => {
					sound.soundOnGlobal = sound.soundOnGlobal ? false : true;
					settings["General"]["SoundOn"] = sound.soundOnGlobal ? "true" : "false";
					saveConfig();
				});
				CreateToggle((toggleSounds), true);
				
				/*
				 toggleAJ = new JSONStorableBool("Auto Jaw Mouth Morph On", etc.autoJawMouthMorphOnGlobal, (bool toggle) => {
					etc.autoJawMouthMorphOnGlobal = etc.autoJawMouthMorphOnGlobal ? false : true;
					settings["General"]["AutoJawMouthMorphOn"] = etc.autoJawMouthMorphOnGlobal ? "true" : "false";
					atom.GetStorableByID("AutoJawMouthMorph").SetBoolParamValue("enabled", !(etc.autoJawMouthMorphOnGlobal));
					saveConfig();
				});
				CreateToggle((toggleAJ), true);
				*/
				
				 toggleBlush = new JSONStorableBool("Blush On", blushing.blushOnGlobal, (bool toggle) => {
					blushing.blushOnGlobal = blushing.blushOnGlobal ? false : true;
					settings["General"]["BlushOn"] = blushing.blushOnGlobal ? "true" : "false";
					if(blushing.blushOnGlobal){
						blushing.update(Time.time);
					}else{
						blushing.isBlushingActive = false;
						blushing.switchFaceDecal("");
					}
					saveConfig();
				});
				CreateToggle((toggleBlush), false);
				
				spacer11 = CreateSpacer(false);
				spacer11.height = 36f;
				
				blushType = new JSONStorableStringChooser("Blush Type", blushing.blushTypes, blushing.blushTypeDefault, "Blush Type", (string type) => {
					blushing.blushAliasLow = blushing.blushMap[type];
					blushing.blushTypeDefault = type;
					settings["General"]["DefaultBlushType"] = blushing.blushTypeDefault;
					if(blushing.blushOnGlobal){
						blushing.isBlushingActive = false;
						blushing.update(Time.time);
					}
					saveConfig();
				});
				RegisterStringChooser(blushType);
				UIDynamicPopup udp = CreateScrollablePopup(blushType, true);
				
				 toggleTears = new JSONStorableBool("Tears On", tears.tearsOnGlobal, (bool toggle) => {
					tears.tearsOnGlobal = tears.tearsOnGlobal ? false : true;
					settings["General"]["TearsOn"] = tears.tearsOnGlobal ? "true" : "false";
					saveConfig();
					tears.refreshTears();
				});
				CreateToggle((toggleTears), false);
				
				 toggleDrool = new JSONStorableBool("Drool On", tears.droolOnGlobal, (bool toggle) => {
					tears.droolOnGlobal = tears.droolOnGlobal ? false : true;
					settings["General"]["DroolOn"] = tears.droolOnGlobal ? "true" : "false";
					saveConfig();
					tears.refreshTears();
				});
				CreateToggle((toggleDrool), false);
				
				tearsType = new JSONStorableStringChooser("Tears Type", tears.tearsTypes, tears.tearsTypeDefault, "Tears Type", (string type) => {
					tears.tearsTypeDefault = type;
					tears.setTearsType(type);
					settings["General"]["DefaultTearsType"] = tears.tearsTypeDefault;
					saveConfig();
				});
				RegisterStringChooser(tearsType);
				UIDynamicPopup udp2 = CreateScrollablePopup(tearsType, true);

				spacer12 = CreateSpacer(true);
				spacer12.height = 2f;

/*				
				 toggleTearsGloss = new JSONStorableBool("Tears Gloss On", tears.isTearsGloss, (bool toggle) => {
					tears.isTearsGloss = tears.isTearsGloss ? false : true;
					settings["General"]["TearsGlossOn"] = tears.isTearsGloss ? "true" : "false";
					saveConfig();
					tears.refreshTears();
				});
				CreateToggle((toggleTearsGloss), false);
*/
				
				 toggleEyecontrol = new JSONStorableBool("Eyes Control On", eyesControl.eyecontrolOnGlobal, (bool toggle) => {
					eyesControl.eyecontrolOnGlobal = eyesControl.eyecontrolOnGlobal ? false : true;
					settings["General"]["EyesControlOn"] = eyesControl.eyecontrolOnGlobal ? "true" : "false";
					if(eyesControl.eyecontrolOnGlobal){
						eyesControl.update(Time.time);
					}else{
						eyesControl.eyeControlBaseSwitch("Player");
						eyesControl.setInitialTransformOfEyeControl();
						eyesControl.isEyecontrolActive = false;
					}
					saveConfig();
				});
				CreateToggle((toggleEyecontrol), false);
				
				 toggleEyecontrolRollup = new JSONStorableBool("Eyes Rollup On", eyesControl.eyecontrolRollupOnGlobal, (bool toggle) => {
					eyesControl.eyecontrolRollupOnGlobal = eyesControl.eyecontrolRollupOnGlobal ? false : true;
					settings["General"]["EyesRollupOn"] = eyesControl.eyecontrolRollupOnGlobal ? "true" : "false";
					saveConfig();
				});
				CreateToggle((toggleEyecontrolRollup), true);
				
				toggleNippleErection = new JSONStorableBool("Nipple Erection On", bodyMorph.nippleErectionOnGlobal, (bool toggle) => {
					bodyMorph.nippleErectionOnGlobal = bodyMorph.nippleErectionOnGlobal ? false : true;
					settings["General"]["NippleErectionOn"] = bodyMorph.nippleErectionOnGlobal ? "true" : "false";
					if(bodyMorph.nippleErectionOnGlobal){
						bodyMorph.update(Time.time);
					}else{
						bodyMorph.stopBodyMorph();
					}
					saveConfig();
				});
				CreateToggle((toggleNippleErection), false);
				
				toggleClitorisErection = new JSONStorableBool("Clitoris Erection On", bodyMorph.clitorisErectionOnGlobal, (bool toggle) => {
					bodyMorph.clitorisErectionOnGlobal = bodyMorph.clitorisErectionOnGlobal ? false : true;
					settings["General"]["ClitorisErectionOn"] = bodyMorph.clitorisErectionOnGlobal ? "true" : "false";
					if(bodyMorph.clitorisErectionOnGlobal){
						bodyMorph.update(Time.time);
					}else{
						bodyMorph.stopBodyMorph();
					}
					saveConfig();
				});
				CreateToggle((toggleClitorisErection), true);
				
				spacer13 = CreateSpacer(false);
				spacer13.height = 16f;
				spacer14 = CreateSpacer(true);
				spacer14.height = 16f;
			}catch (Exception e){
				loggerE(e);
			}
		}

		public void setupUIMorphsE(){
			try{
				int num = expression.mmEmotion.ms.Count;
				string tag = "SettingForEmotions";
				togglesE = new JSONStorableBool[num];
				stringE  = new JSONStorableString[num];
				testE    = new JSONStorableFloat[num];
				limitE   = new JSONStorableFloat[num];
				spacerE  = new UIDynamic[num];

				int i = 0;
				string FieldName = "";
				string explanation = "";
				string expPrefix = "MorphName:";
				string expSuffix = "For this morph, you can\n - toggle enable/disable\n - limit strength\n - test morph";
				foreach(var kvp in expression.mmEmotion.ms){
					FieldName = "TEXT:" + kvp.Key;
					explanation = expPrefix + "\n  " + kvp.Key + "\n" + expSuffix;
					stringE[i] = new JSONStorableString(FieldName, explanation);
					CreateTextField(stringE[i], false);
					
					spacerE[i] = CreateSpacer(false);
					spacerE[i].height = 105f;
					
					FieldName = "Enabled";
					togglesE[i] = new JSONStorableBool(FieldName, expression.mmEmotion.ms[kvp.Key].enabled, (bool toggle) => {
						expression.mmEmotion.ms[kvp.Key].enabled = expression.mmEmotion.ms[kvp.Key].enabled ? false : true;
						settings[tag][kvp.Key]["Enabled"] = expression.mmEmotion.ms[kvp.Key].enabled.ToString();
						expression.mmEmotion.repackEnabledMorphs();
						saveConfig();
					});
					CreateToggle((togglesE[i]), true);
					
					FieldName = "Morph Strength Limit";
					limitE[i] = new JSONStorableFloat(FieldName, expression.mmEmotion.ms[kvp.Key].limit, (float value) => {
						expression.mmEmotion.ms[kvp.Key].limit = (float)value;
						settings[tag][kvp.Key]["Limit"] = value.ToString();
						saveConfig();
					}, 0f, 1.0f);
					CreateSlider(limitE[i], true);
					
					FieldName = "Morph Test";
					testE[i] = new JSONStorableFloat(FieldName, 0f, (float value) => {						
						expression.mmEmotion.ms[kvp.Key].morph.SetValue(value);
					}, 0f, 1.0f);
					CreateSlider(testE[i], true);					
					
					i++;
				}
			}catch (Exception e){
				loggerE(e);
			}
		}
			
		public void setupUIMorphsO(){
			try{
				int num = expression.mmOrgasm.ms.Count;
				string tag = "SettingForOrgasm";
				togglesO = new JSONStorableBool[num];
				stringO  = new JSONStorableString[num];
				testO    = new JSONStorableFloat[num];
				limitO   = new JSONStorableFloat[num];
				spacerO  = new UIDynamic[num];

				int i = 0;
				string FieldName = "";
				string explanation = "";
				string expPrefix = "MorphName:";
				string expSuffix = "For this morph, you can\n - toggle enable/disable\n - limit strength\n - test morph";
				foreach(var kvp in expression.mmOrgasm.ms){
					FieldName = "TEXT:" + kvp.Key;
					explanation = expPrefix + "\n  " + kvp.Key + "\n" + expSuffix;
					stringO[i] = new JSONStorableString(FieldName, explanation);
					CreateTextField(stringO[i], false);
					
					spacerO[i] = CreateSpacer(false);
					spacerO[i].height = 105f;
					
					FieldName = "Enabled";
					togglesO[i] = new JSONStorableBool(FieldName, expression.mmOrgasm.ms[kvp.Key].enabled, (bool toggle) => {
						expression.mmOrgasm.ms[kvp.Key].enabled = expression.mmOrgasm.ms[kvp.Key].enabled ? false : true;
						settings[tag][kvp.Key]["Enabled"] = expression.mmOrgasm.ms[kvp.Key].enabled.ToString();
						expression.mmOrgasm.repackEnabledMorphs();
						saveConfig();
					});
					CreateToggle((togglesO[i]), true);
					
					FieldName = "Morph Strength Limit";
					limitO[i] = new JSONStorableFloat(FieldName, expression.mmOrgasm.ms[kvp.Key].limit, (float value) => {
						expression.mmOrgasm.ms[kvp.Key].limit = (float)value;
						settings[tag][kvp.Key]["Limit"] = value.ToString();
						saveConfig();
					}, 0f, 1.0f);
					CreateSlider(limitO[i], true);
					
					FieldName = "Morph Test";
					testO[i] = new JSONStorableFloat(FieldName, 0f, (float value) => {						
						expression.mmOrgasm.ms[kvp.Key].morph.SetValue(value);
					}, 0f, 1.0f);
					CreateSlider(testO[i], true);					
					
					i++;
				}
			}catch (Exception e){
				loggerE(e);
			}
		}
			
		public void setupUISliders1(){
			try{
				jfBlush = new JSONStorableFloat("Blush Duration Time", blushing.blushClosingTimeOffset, (float value) => {
					blushing.blushClosingTimeOffset = value;
					settings["General"]["BlushDurationTime"] = value.ToString();
					saveConfig();
				}, 0f, 60f);
				CreateSlider(jfBlush, false);

				 jfTears = new JSONStorableFloat("Tears Duration Time", tears.tearsClosingTimeOffset, (float value) => {
					tears.tearsClosingTimeOffset = value;
					settings["General"]["TearsDurationTime"] = value.ToString();
					saveConfig();
				}, 0f, 60f);
				CreateSlider(jfTears, true);
				
				jfBodyMorph = new JSONStorableFloat("BodyMorph Duration Time", bodyMorph.bodyMorphClosingTimeOffset, (float value) => {
					bodyMorph.bodyMorphClosingTimeOffset = value;
					settings["General"]["BodyMorphDurationTime"] = value.ToString();
					saveConfig();
				}, 0f, 60f);
				CreateSlider(jfBodyMorph, false);
				
				 jfEyecontrol = new JSONStorableFloat("Eyes Control interval Time", eyesControl.eyecontrolClosingTimeInterval, (float value) => {
					eyesControl.eyecontrolClosingTimeInterval = value;
					settings["General"]["EyesControlDurationTime"] = value.ToString();
					saveConfig();
				}, 0f, 60f);
				CreateSlider(jfEyecontrol, true);			
				
				 jfExpressionUp = new JSONStorableFloat("Expression Up Time", expression.expressionClosingTimeOffset1, (float value) => {
					expression.expressionClosingTimeOffset1 = value;
					settings["General"]["ExpressionUpTime"] = value.ToString();
					saveConfig();
				}, 0f, 1f);
				CreateSlider(jfExpressionUp, false);
				
				 jfExpressionDown = new JSONStorableFloat("Expression Down Time", expression.expressionClosingTimeOffset2, (float value) => {
					expression.expressionClosingTimeOffset2 = value;
					settings["General"]["ExpressionDownTime"] = value.ToString();
					saveConfig();
				}, 0f, 60f);
				CreateSlider(jfExpressionDown, true);
			
				 jfExpressionOffsetMin = new JSONStorableFloat("Expression Duration Time Offset(Min)", expression.expressionRandomOffsetMin, (float value) => {
					expression.expressionRandomOffsetMin = value;
					settings["General"]["ExpressionRandomOffsetMin"] = value.ToString();
					saveConfig();
				}, 0f, 60f);
				CreateSlider(jfExpressionOffsetMin, false);
				
				 jfExpressionOffsetMax = new JSONStorableFloat("Expression Duration Time Offset(Max)", expression.expressionRandomOffsetMax, (float value) => {
					expression.expressionRandomOffsetMax = value;
					settings["General"]["ExpressionRandomOffsetMax"] = value.ToString();
					saveConfig();
				}, 0f, 60f);
				CreateSlider(jfExpressionOffsetMax, true);
				
				 jfExpressionStrength = new JSONStorableFloat("Expression Strength", expression.expressionStrengthBase, (float value) => {
					expression.expressionStrengthBase = value;
					settings["General"]["ExpressionStrengthBase"] = value.ToString();
					saveConfig();
				}, 0f, 60f);
				CreateSlider(jfExpressionStrength, false);

				jfBendBackRatio = new JSONStorableFloat("Bending Backward Ratio", belly.bendBackwardThreshold, (float value) => {
					belly.bendBackwardThreshold = (float)value;
					settings["Orgasm"]["BendingBackwardRatio"] = value.ToString();
					saveConfig();
				}, 0f, 1000f);
				CreateSlider(jfBendBackRatio, true);

				jfBendBackStrength = new JSONStorableFloat("Bending Backward Strength", belly.targetStrength["headControl"], (float value) => {
					belly.targetStrength["headControl"] = (float)value;
					settings["General"]["BendingBackStrength"] = value.ToString();
					saveConfig();
				}, -60f, 60f);
				CreateSlider(jfBendBackStrength, true);

				jfBreathStrength = new JSONStorableFloat("Breath Strength", belly.targetStrength["chestControl"], (float value) => {
					belly.targetStrength["chestControl"] = (float)value;
					settings["General"]["BreathStrength"] = value.ToString();
					saveConfig();
				}, -60f, 60f);
				CreateSlider(jfBreathStrength, false);

				spacer21 = CreateSpacer(false);
				spacer21.height = 16f;
				spacer22 = CreateSpacer(true);
				spacer22.height = 16f;
			}catch (Exception e){
				loggerE(e);
			}
		}

		public void setupUISliders2(){
			try{
				jfThresholds = new JSONStorableFloat[6];
				uiClothesResetButtons = new UIDynamicButton[6];
				uiClothesSetButtons   = new UIDynamicButton[6];
				uiClothesTestButtons  = new UIDynamicButton[6];
				uiClothesDebugButtons = new UIDynamicButton[6];
				uiClothesStrings      = new JSONStorableString[6];
				uiClothesSpacers      = new UIDynamic[6];

				string expPrefix = "Settings for each Stage Threshold";
				string expSuffix = "For each stage, you can set the threshold and the clothes you want the character to equip with this menu.\nIn this Plug-in, tears and drool are implemented as clothes.";
				
				uiUndressAllButton = CreateButton("Undress All", false);
				uiUndressAllButton.button.onClick.AddListener(() =>{
					tears.cm.undressAll();
				});
				
				spacer34 = CreateSpacer(true);
				spacer34.height = 52f;
				
				for(int i = 0; i < 6; i++){
					int num = i;
					string FieldName = "TEXT_UiClothes" + num;
					string explanation = expPrefix + num + ":\n" + expSuffix;
					uiClothesStrings[num] = new JSONStorableString(FieldName, explanation);
					CreateTextField(uiClothesStrings[num], false);
					
					jfThresholds[num] = new JSONStorableFloat("Stage" + num + " Threshold", (float)stageThreshold[num], (float value) => {
						stageThreshold[num] = (int)value;
						settings["StageThreshold"][("Threshold" + num)]["Value"] = stageThreshold[num].ToString();
						saveConfig();
					}, 0f, 1000f);
					CreateSlider(jfThresholds[num], true);
					
					uiClothesSetButtons[num] = CreateButton("Set Stage" + num + " clothes", true);
					uiClothesSetButtons[num].button.onClick.AddListener(() =>{
						tears.cm.setClothes(num);
						List<string> tmpClothUids = tears.cm.getClothUids(num);
						for(int j = 0; j < tmpClothUids.Count; j++){
							settings["StageThreshold"][("Threshold" + num)]["Clothes"][j] = tmpClothUids[j];
						}
						saveConfig();
					});	
					
					uiClothesResetButtons[num] = CreateButton("Reset Stage" + num + " clothes", true);
					uiClothesResetButtons[num].button.onClick.AddListener(() =>{
						tears.cm.clearClothes(num);
						settings["StageThreshold"][("Threshold" + num)]["Clothes"] = new JSONArray();						
						saveConfig();
					});	
					
					uiClothesTestButtons[num] = CreateButton("Test Stage" + num + " clothes", true);
					uiClothesTestButtons[num].button.onClick.AddListener(() =>{
						tears.cm.equipClothes(num);
					});	
					
					uiClothesSpacers[num] = CreateSpacer(false);
					uiClothesSpacers[num].height = 36f;
					
					uiClothesDebugButtons[num] = CreateButton("Show Clothes for Stage" + num, false);
					uiClothesDebugButtons[num].button.onClick.AddListener(() =>{
						SuperController.LogMessage("Clothes for Stage" + num + ":");
						tears.cm.showClothes(num);
					});	
					
				}
				
				spacer31 = CreateSpacer(true);
				spacer31.height = 120f;
				spacer32 = CreateSpacer(false);
				spacer32.height = 16f;
				spacer33 = CreateSpacer(true);
				spacer33.height = 16f;
			}catch (Exception e){
				loggerE(e);
			}
		}

		public void setupUISound(){
			try{
				string FieldName = "TEXT:Sound";
				string explanation = "You can change or test the pitch of the voice in this menu.";
				stringS = new JSONStorableString(FieldName, explanation);
				CreateTextField(stringS, false);

				spacer51 = CreateSpacer(false);
				spacer51.height = 102f;

				jfPitch = new JSONStorableFloat("Sound Pitch", sound.headAudioSource.pitch, (float value) => {
					sound.headAudioSource.pitch = value;
					settings["General"]["Pitch"] = value.ToString();
					saveConfig();
				}, 0.8f, 1.5f);
				CreateSlider(jfPitch, true);
		
				soundTestButtonBreath = CreateButton("Sound Test: Breath", true);
				soundTestButtonBreath.button.onClick.AddListener(() =>{
					sound.playSound("breath", Time.time);
				});	
		
				soundTestButtonMoan = CreateButton("Sound Test: Moan", true);
				soundTestButtonMoan.button.onClick.AddListener(() =>{
					sound.playSound("moan", Time.time);
				});	
		
				soundTestButtonClimax = CreateButton("Sound Test: Climax", true);
				soundTestButtonClimax.button.onClick.AddListener(() =>{
					sound.playSound("climax", Time.time);
				});	
		
			}catch (Exception e){
				loggerE(e);
			}
		}
		public void setupUISliders3(){
			try{
//				spacer12 = CreateSpacer(true);
//				spacer12.height = 120f;

#if (EBATWO)
				jfOStrengthMax = new JSONStorableFloat("Orgasm Strength Max", (float)orgasm.orgasmStrengthMax, (float value) => {
					orgasm.orgasmStrengthMax = (float)value;
					settings["Orgasm"]["OrgasmStrength"] = value.ToString();
					saveConfig();
				}, 0f, 5.0f);
				CreateSlider(jfOStrengthMax, false);

				jfAStrengthMax = new JSONStorableFloat("Afterglow Strength Max", (float)orgasm.afterglowStrengthMax, (float value) => {
					orgasm.afterglowStrengthMax = (float)value;
					settings["Orgasm"]["AfterglowStrength"] = value.ToString();
					saveConfig();
				}, 0f, 5.0f);
				CreateSlider(jfAStrengthMax, true);

				jfOWCountBigMax = new JSONStorableFloat("Orgasm Wave Big Cycle Count Max", (float)orgasm.orgasmWaveCycleBigMax, (float value) => {
					orgasm.orgasmWaveCycleBigMax = (int)value;
					settings["Orgasm"]["WaveBigCycleMax"] = value.ToString();
					saveConfig();
				}, 0f, 1000f);
				CreateSlider(jfOWCountBigMax, false);

				jfOWCountBigMin = new JSONStorableFloat("Orgasm Wave Big Cycle Count Min", (float)orgasm.orgasmWaveCycleBigMin, (float value) => {
					orgasm.orgasmWaveCycleBigMin = (int)value;
					settings["Orgasm"]["WaveBigCycleMin"] = value.ToString();
					saveConfig();
				}, 0f, 1000f);
				CreateSlider(jfOWCountBigMin, true);

				jfOWCountSmallMax = new JSONStorableFloat("Orgasm Wave Small Cycle Count Max", (float)orgasm.orgasmWaveCycleSmallMax, (float value) => {
					orgasm.orgasmWaveCycleSmallMax = (int)value;
					settings["Orgasm"]["WaveSmallCycleMax"] = value.ToString();
					saveConfig();
				}, 0f, 1000f);
				CreateSlider(jfOWCountSmallMax, false);

				jfOWCountSmallMin = new JSONStorableFloat("Orgasm Wave Small Cycle Count Min", (float)orgasm.orgasmWaveCycleSmallMin, (float value) => {
					orgasm.orgasmWaveCycleSmallMin = (int)value;
					settings["Orgasm"]["WaveSmallCycleMin"] = value.ToString();
					saveConfig();
				}, 0f, 1000f);
				CreateSlider(jfOWCountSmallMin, true);
				
				jfOACountBigMax = new JSONStorableFloat("Orgasm Afterglow Big Cycle Count Max", (float)orgasm.orgasmAfterglowCycleBigMax, (float value) => {
					orgasm.orgasmAfterglowCycleBigMax = (int)value;
					settings["Orgasm"]["AfterglowBigCycleMax"] = value.ToString();
					saveConfig();
				}, 0f, 1000f);
				CreateSlider(jfOACountBigMax, false);

				jfOACountBigMin = new JSONStorableFloat("Orgasm Afterglow Big Cycle Count Min", (float)orgasm.orgasmAfterglowCycleBigMin, (float value) => {
					orgasm.orgasmAfterglowCycleBigMin = (int)value;
					settings["Orgasm"]["AfterglowBigCycleMin"] = value.ToString();
					saveConfig();
				}, 0f, 1000f);
				CreateSlider(jfOACountBigMin, true);

				jfOACountSmallMax = new JSONStorableFloat("Orgasm Afterglow Small Cycle Count Max", (float)orgasm.orgasmAfterglowCycleSmallMax, (float value) => {
					orgasm.orgasmAfterglowCycleSmallMax = (int)value;
					settings["Orgasm"]["AfterglowSmallCycleMax"] = value.ToString();
					saveConfig();
				}, 0f, 1000f);
				CreateSlider(jfOACountSmallMax, false);

				jfOACountSmallMin = new JSONStorableFloat("Orgasm Afterglow Small Cycle Count Min", (float)orgasm.orgasmAfterglowCycleSmallMin, (float value) => {
					orgasm.orgasmAfterglowCycleSmallMin = (int)value;
					settings["Orgasm"]["AfterglowSmallCycleMin"] = value.ToString();
					saveConfig();
				}, 0f, 1000f);
				CreateSlider(jfOACountSmallMin, true);

				jfOWIntervalBigMax = new JSONStorableFloat("Orgasm Wave Big Cycle Interval Max", (float)orgasm.orgasmWaveCycleIntervalBigMax, (float value) => {
					orgasm.orgasmWaveCycleIntervalBigMax = (float)value;
					settings["Orgasm"]["WaveBigCycleIntervalMax"] = value.ToString();
					saveConfig();
				}, 0f, 1000f);
				CreateSlider(jfOWIntervalBigMax, false);

				jfOWIntervalSmallMax = new JSONStorableFloat("Orgasm Wave Small Cycle Interval Max", (float)orgasm.orgasmWaveCycleIntervalSmallMax, (float value) => {
					orgasm.orgasmWaveCycleIntervalSmallMax = (float)value;
					settings["Orgasm"]["WaveSmallCycleIntervalMax"] = value.ToString();
					saveConfig();
				}, 0f, 1000f);
				CreateSlider(jfOWIntervalSmallMax, true);

				jfOAIntervalBigMax = new JSONStorableFloat("Orgasm Afterglow Big Cycle Interval Max", (float)orgasm.orgasmAfterglowCycleIntervalBigMax, (float value) => {
					orgasm.orgasmAfterglowCycleIntervalBigMax = (float)value;
					settings["Orgasm"]["AfterglowBigCycleIntervalMax"] = value.ToString();
					saveConfig();
				}, 0f, 1000f);
				CreateSlider(jfOAIntervalBigMax, false);

				jfOAIntervalSmallMax = new JSONStorableFloat("Orgasm Afterglow Small Cycle Interval Max", (float)orgasm.orgasmAfterglowCycleIntervalSmallMax, (float value) => {
					orgasm.orgasmAfterglowCycleIntervalSmallMax = (float)value;
					settings["Orgasm"]["AfterglowSmallCycleIntervalMax"] = value.ToString();
					saveConfig();
				}, 0f, 1000f);
				CreateSlider(jfOAIntervalSmallMax, true);

				spacer41 = CreateSpacer(true);
				spacer41.height = 16f;
				spacer42 = CreateSpacer(true);
				spacer42.height = 16f;
#endif
			}catch (Exception e){
				loggerE(e);
			}
		}
		
		public void removeConfigMenu(){
			if(toggleDebug != null){
				removeUIActionButtons();
			}
			removeUIExpandToggleMenu();
			removeUIExpandSoundMenu();
			removeUIExpandMorphEMenu();
			removeUIExpandMorphOMenu();
			removeUIExpandSlider1Menu();
			removeUIExpandSlider2Menu();
			if(expandToggleMenu){
				removeUIToggleButtons();
			}
			if(expandSoundMenu){
				removeUISound();
			}
			if(expandMorphEMenu){
				removeUIMorphsE();
			}
			if(expandMorphOMenu){
				removeUIMorphsO();
			}
			if(expandSlider1Menu){
				removeUISliders1();
			}
			if(expandSlider2Menu){
				removeUISliders2();
			}
#if (EBATWO)
			removeUIExpandSlider3Menu();
			if(expandSlider3Menu){
				removeUISliders3();
			}
#endif
		}

		public void removeConfigMenuForce(){
			if(toggleDebug != null){
				removeUIActionButtons();
			}
			if(expandToggleButton != null){
				removeUIExpandToggleMenu();
			}
			if(toggleAE != null){
				removeUIToggleButtons();
			}
			if(expandMorphEButton != null){
				removeUIExpandMorphEMenu();
			}
			if(togglesE[0] != null){
				removeUIMorphsE();
			}
			if(expandMorphEButton != null){
				removeUIExpandMorphOMenu();
			}
			if(togglesO[0] != null){
				removeUIMorphsO();
			}
			if(expandSlider1Button != null){
				removeUIExpandSlider1Menu();
			}
			if(jfBlush != null){
				removeUISliders1();
			}
			if(expandSlider2Button != null){
				removeUIExpandSlider2Menu();
			}
			if(jfThresholds[0] != null){
				removeUISliders2();
			}
#if (EBATWO)
			if(expandSlider3Button != null){
				removeUIExpandSlider3Menu();
			}
			if(jfOWCountBigMax != null){
				removeUISliders3();
			}
#endif
		}
		
		public void removeUIExpandToggleMenu(){
			try{
				RemoveButton(expandToggleButton);
				RemoveButton(collapseToggleButton);	
				RemoveSpacer(spacerA1);
				RemoveSpacer(spacerA2);
			}catch (Exception e){
				loggerE(e);
			}			
		}
		
		public void removeUIExpandMorphEMenu(){
			try{
				RemoveButton(expandMorphEButton);
				RemoveButton(collapseMorphEButton);				
				RemoveSpacer(spacerF1);
				RemoveSpacer(spacerF2);
			}catch (Exception e){
				loggerE(e);
			}			
		}
		
		public void removeUIExpandMorphOMenu(){
			try{
				RemoveButton(expandMorphOButton);
				RemoveButton(collapseMorphOButton);				
				RemoveSpacer(spacerB1);
				RemoveSpacer(spacerB2);
			}catch (Exception e){
				loggerE(e);
			}			
		}
		
		public void removeUIExpandSlider1Menu(){
			try{
				RemoveButton(expandSlider1Button);
				RemoveButton(collapseSlider1Button);				
				RemoveSpacer(spacerC1);
				RemoveSpacer(spacerC2);
			}catch (Exception e){
				loggerE(e);
			}			
		}
		
		public void removeUIExpandSlider2Menu(){
			try{
				RemoveButton(expandSlider2Button);
				RemoveButton(collapseSlider2Button);				
				RemoveSpacer(spacerD1);
				RemoveSpacer(spacerD2);
			}catch (Exception e){
				loggerE(e);
			}			
		}
		
		public void removeUIExpandSlider3Menu(){
			try{
				RemoveButton(expandSlider3Button);
				RemoveButton(collapseSlider3Button);				
				RemoveSpacer(spacerE1);
				RemoveSpacer(spacerE2);
			}catch (Exception e){
				loggerE(e);
			}			
		}
		
		public void removeUIExpandSoundMenu(){
			try{
				RemoveButton(expandSoundButton);
				RemoveButton(collapseSoundButton);				
				RemoveSpacer(spacerG1);
				RemoveSpacer(spacerG2);
			}catch (Exception e){
				loggerE(e);
			}			
		}
		
		public void removeUIActionButtons(){
			try{
				RemoveToggle(toggleDebug);				
				RemoveButton(resetButton);
				RemoveButton(copySoundButton);
#if (EBATWO)
				RemoveButton(orgasmButton);
				RemoveButton(stopOrgasmButton);
#endif
				RemoveSpacer(spacer01);
				RemoveSpacer(spacer02);
				RemoveSpacer(spacer03);
			}catch (Exception e){
				loggerE(e);
			}
		}
		
		public void removeUIToggleButtons(){
			try{
				RemoveToggle(toggleAE);
				RemoveToggle(toggleAB);
//				RemoveToggle(toggleAJ);
				RemoveToggle(toggleSounds);
				RemoveSpacer(spacer11);
				RemovePopup(blushType);
				RemoveSpacer(spacer12);
				RemoveToggle(toggleBlush);
				RemoveToggle(toggleTears);
				RemoveToggle(toggleDrool);
				RemovePopup(tearsType);
//				RemoveToggle(toggleTearsGloss);
				RemoveToggle(toggleExpression);
				RemoveToggle(toggleEyecontrol);
				RemoveToggle(toggleEyecontrolRollup);
				RemoveToggle(toggleNippleErection);
				RemoveToggle(toggleClitorisErection);
				
				RemoveSpacer(spacer13);
				RemoveSpacer(spacer14);
			}catch (Exception e){
				loggerE(e);
			}
		}
		
		public void removeUISound(){
			try{
				RemoveTextField(stringS);
				RemoveSlider(jfPitch);
				RemoveSpacer(spacer51);
				RemoveButton(soundTestButtonBreath);
				RemoveButton(soundTestButtonMoan);
				RemoveButton(soundTestButtonClimax);
			}catch (Exception e){
				loggerE(e);
			}			
		}
		
		public void removeUIMorphsO(){
			try{
				for(int i = 0; i < stringO.Length; i++){
					RemoveToggle(togglesO[i]);
					RemoveTextField(stringO[i]);
					RemoveSlider(limitO[i]);
					RemoveSlider(testO[i]);
					RemoveSpacer(spacerO[i]);
				}
			}catch (Exception e){
				loggerE(e);
			}
		}
		
		public void removeUIMorphsE(){
			try{
				for(int i = 0; i < stringE.Length; i++){
					RemoveToggle(togglesE[i]);
					RemoveTextField(stringE[i]);
					RemoveSlider(limitE[i]);
					RemoveSlider(testE[i]);
					RemoveSpacer(spacerE[i]);
				}
			}catch (Exception e){
				loggerE(e);
			}
		}
		
		public void removeUISliders1(){
			try{
				RemoveSlider(jfBlush);
				RemoveSlider(jfTears);
				RemoveSlider(jfBodyMorph);
				RemoveSlider(jfEyecontrol);
				RemoveSlider(jfExpressionUp);
				RemoveSlider(jfExpressionDown);
				RemoveSlider(jfExpressionOffsetMin);
				RemoveSlider(jfExpressionOffsetMax);
				RemoveSlider(jfExpressionStrength);
				RemoveSlider(jfBendBackRatio);
				RemoveSlider(jfBendBackStrength);
				RemoveSlider(jfBreathStrength);
				RemoveSpacer(spacer21);
				RemoveSpacer(spacer22);
			}catch (Exception e){
				loggerE(e);
			}
		}
		
		public void removeUISliders2(){
			try{
				RemoveButton(uiUndressAllButton);
				RemoveSpacer(spacer34);
				for(int i = 0; i < 6; i++){
					RemoveTextField(uiClothesStrings[i]);
					RemoveSlider(jfThresholds[i]);
					RemoveButton(uiClothesResetButtons[i]);
					RemoveButton(uiClothesSetButtons[i]);
					RemoveButton(uiClothesTestButtons[i]);
					RemoveSpacer(uiClothesSpacers[i]);
					RemoveButton(uiClothesDebugButtons[i]);
				}
		
				RemoveSpacer(spacer31);
				RemoveSpacer(spacer32);
				RemoveSpacer(spacer33);
			}catch (Exception e){
				loggerE(e);
			}
		}
		
		public void removeUISliders3(){
			try{		
//				RemoveSpacer(spacer12);
				RemoveSlider(jfOWCountBigMax);
				RemoveSlider(jfOWCountBigMin);
				RemoveSlider(jfOWCountSmallMax);
				RemoveSlider(jfOWCountSmallMin);
				RemoveSlider(jfOACountBigMax);
				RemoveSlider(jfOACountBigMin);
				RemoveSlider(jfOACountSmallMax);
				RemoveSlider(jfOACountSmallMin);
				RemoveSlider(jfOWIntervalBigMax);
				RemoveSlider(jfOWIntervalSmallMax);
				RemoveSlider(jfOAIntervalBigMax);
				RemoveSlider(jfOAIntervalSmallMax);
				RemoveSlider(jfOStrengthMax);
				RemoveSlider(jfAStrengthMax);
				
				RemoveSpacer(spacer41);
				RemoveSpacer(spacer42);
			}catch (Exception e){
				loggerE(e);
			}				
		}
		
		#region ui var
		protected bool expandToggleMenu = false;
		protected bool expandMorphOMenu = false;
		protected bool expandMorphEMenu = false;
		protected bool expandSlider1Menu = false;
		protected bool expandSlider2Menu = false;
		protected bool expandSlider3Menu = false;
		protected bool expandSoundMenu = false;
		
		protected UIDynamic spacer01;
		protected UIDynamic spacer02;
		protected UIDynamic spacer03;
		protected UIDynamic spacer11;
		protected UIDynamic spacer12;
		protected UIDynamic spacer13;
		protected UIDynamic spacer14;
		protected UIDynamic spacer21;
		protected UIDynamic spacer22;
		protected UIDynamic spacer9;
		protected UIDynamic spacer31;
		protected UIDynamic spacer32;
		protected UIDynamic spacer33;
		protected UIDynamic spacer34;
		protected UIDynamic spacer41;
		protected UIDynamic spacer42;
		protected UIDynamic spacer51;

		protected UIDynamic spacerA1;
		protected UIDynamic spacerA2;
		protected UIDynamic spacerB1;
		protected UIDynamic spacerB2;
		protected UIDynamic spacerC1;
		protected UIDynamic spacerC2;
		protected UIDynamic spacerD1;
		protected UIDynamic spacerD2;
		protected UIDynamic spacerE1;
		protected UIDynamic spacerE2;
		protected UIDynamic spacerF1;
		protected UIDynamic spacerF2;
		protected UIDynamic spacerG1;
		protected UIDynamic spacerG2;

		protected UIDynamicButton expandToggleButton;
		protected UIDynamicButton expandMorphOButton;
		protected UIDynamicButton expandMorphEButton;
		protected UIDynamicButton expandSlider1Button;
		protected UIDynamicButton expandSlider2Button;
		protected UIDynamicButton expandSlider3Button;
		protected UIDynamicButton expandSoundButton;
		protected UIDynamicButton collapseToggleButton;
		protected UIDynamicButton collapseMorphOButton;
		protected UIDynamicButton collapseMorphEButton;
		protected UIDynamicButton collapseSlider1Button;
		protected UIDynamicButton collapseSlider2Button;
		protected UIDynamicButton collapseSlider3Button;
		protected UIDynamicButton collapseSoundButton;

		protected JSONStorableString stringS;
		protected UIDynamicButton soundTestButtonBreath;
		protected UIDynamicButton soundTestButtonMoan;
		protected UIDynamicButton soundTestButtonClimax;

		public JSONStorableBool[] togglesO;
		public JSONStorableString[] stringO;
		public JSONStorableFloat[] testO;
		public JSONStorableFloat[] limitO;
		public UIDynamic[] spacerO;
		
		public JSONStorableBool[] togglesE;
		public JSONStorableString[] stringE;
		public JSONStorableFloat[] testE;
		public JSONStorableFloat[] limitE;
		public UIDynamic[] spacerE;		
		
		protected JSONStorableStringChooser blushType;
		protected JSONStorableStringChooser tearsType;
		protected JSONStorableBool toggleAE;
		protected JSONStorableBool toggleAB;
		protected JSONStorableBool toggleAJ;
		protected JSONStorableBool toggleSounds;
		protected JSONStorableBool toggleBlush;
		protected JSONStorableBool toggleTears;
		protected JSONStorableBool toggleNippleErection;
		protected JSONStorableBool toggleClitorisErection;
		protected JSONStorableBool toggleDrool;
		protected JSONStorableBool toggleTearsGloss;
		protected JSONStorableBool toggleExpression;
		protected JSONStorableBool toggleEyecontrol;
		protected JSONStorableBool toggleEyecontrolRollup;
		protected JSONStorableBool toggleDebug;
		
		protected JSONStorableFloat jfPitch;
		protected JSONStorableFloat jfBlush;
		protected JSONStorableFloat jfTears;
		protected JSONStorableFloat jfBodyMorph;
		protected JSONStorableFloat jfEyecontrol;
		protected JSONStorableFloat jfExpressionUp;
		protected JSONStorableFloat jfExpressionDown;
		protected JSONStorableFloat jfExpressionOffsetMin;
		protected JSONStorableFloat jfExpressionOffsetMax;
		protected JSONStorableFloat jfExpressionStrength;
		
		protected JSONStorableFloat[] jfThresholds;
		protected UIDynamicButton[] uiClothesSetButtons;
		protected UIDynamicButton[] uiClothesTestButtons;
		protected UIDynamicButton[] uiClothesResetButtons;
		protected UIDynamicButton[] uiClothesDebugButtons;
		protected JSONStorableString[] uiClothesStrings;
		protected UIDynamic[] uiClothesSpacers;
		protected UIDynamicButton uiUndressAllButton;
		
		protected UIDynamicButton resetButton;
		protected UIDynamicButton copySoundButton;
		protected UIDynamicButton orgasmButton;
		protected UIDynamicButton stopOrgasmButton;
		
		protected JSONStorableFloat jfBendBackRatio;		
		protected JSONStorableFloat jfBendBackStrength;		
		protected JSONStorableFloat jfBreathStrength;		
		protected JSONStorableFloat jfOStrengthMax;		
		protected JSONStorableFloat jfAStrengthMax;		
		protected JSONStorableFloat jfOWCountBigMax;
		protected JSONStorableFloat jfOWCountBigMin;
		protected JSONStorableFloat jfOWCountSmallMax;
		protected JSONStorableFloat jfOWCountSmallMin;
		protected JSONStorableFloat jfOACountBigMax;
		protected JSONStorableFloat jfOACountBigMin;
		protected JSONStorableFloat jfOACountSmallMax;
		protected JSONStorableFloat jfOACountSmallMin;
		protected JSONStorableFloat jfOWIntervalBigMax;
		protected JSONStorableFloat jfOWIntervalSmallMax;
		protected JSONStorableFloat jfOAIntervalBigMax;
		protected JSONStorableFloat jfOAIntervalSmallMax;
		#endregion

	}
}
