#region using
using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
#endregion

namespace Cotyounoyume{
	public class ControllerBackup{
		public Vector3 v3pos;
		public Quaternion v3rot;
		public string positionState;
		public string rotationState;
		
		public ControllerBackup(){
			v3pos = new Vector3(0f, 0f, 0f);
			v3rot = new Quaternion(0f, 0f, 0f, 0f);
			positionState = "Off";
			rotationState = "Off";
		}
		
		public void backUp(FreeControllerV3 target){
			try{
				Vector3 tarPos = target.transform.position;
				v3pos.Set(tarPos.x, tarPos.y, tarPos.z);
				Quaternion tarRot = target.transform.rotation;
				v3rot.Set(tarRot.x, tarRot.y, tarRot.z, tarRot.w);
				positionState = target.currentPositionState.ToString();
				rotationState = target.currentRotationState.ToString();
			}catch(Exception e){
				SuperController.LogMessage("error in ControllerBackup.backUp:" + e);
			}
		}
		
		public FreeControllerV3 recovery(FreeControllerV3 target){
			try{
				target.SetPositionStateFromString(positionState);
				target.SetRotationStateFromString(rotationState);
				target.transform.SetPositionAndRotation(v3pos, v3rot);

				return target;
			}catch(Exception e){
				SuperController.LogMessage("error in ControllerBackup.recovery:" + e);
				return null;
			}
		}

		public FreeControllerV3 recoveryPositionOnly(FreeControllerV3 target){
			try{
				target.transform.SetPositionAndRotation(v3pos, v3rot);

				return target;
			}catch(Exception e){
				SuperController.LogMessage("error in ControllerBackup.recovery:" + e);
				return null;
			}
		}

	}
	
	public class CycleManager
	{
		public float opening;
		public float closing1;
		public float closing2;
		public int state;
		
		public CycleManager(){
			
		}
		
		public void setTimer(float openingTime, float closingTIme1, float closingTIme2){
			opening = openingTime;
			closing1 = closingTIme1;
			closing2 = closingTIme2;
		}
		
		public float getProgress(float time){
			float delta;
			float result;
			if(time >= closing2){
				result = 0f;
			}else if(time < closing1){
				delta = time - opening;
				result = delta / (closing1 - opening);
			}else if(time < closing2){
				delta = closing2 - time;
				result = delta / (closing2 - closing1);
			}else{
				result = -1f;
			}
			return result;
		}

		public int getState(float time){
			if(time >= closing2){
				state = 0;
			}else if(time < closing1){
				state = 1;
			}else if(time < closing2){
				state = 2;
			}
			return state;
		}
	}
	
	
	
	public class Belly
	{
		#region common or UI Properties
		public Atom atom;
		public EBATCore core;
		public bool disableLog;
		#endregion
		
		#region bellyControl
		protected Dictionary<string, Quaternion> initialQuat;
		protected Dictionary<string, Quaternion> targetQuat;
		protected Dictionary<string, float> progress;
		protected Dictionary<string, CycleManager> timing;
		protected Dictionary<string, FreeControllerV3> ctrl;
		protected Dictionary<string, ControllerBackup> backup;
		protected Dictionary<string, bool> isActive;

		protected string[] keyBreath = {
			"chestControl"
		};
		protected string[] keyOrgasm = {
			"abdomenControl", "abdomen2Control", "pelvisControl", "hipControl"
			, "lThighControl", "lKneeControl", "lFootControl"
			, "rThighControl", "rKneeControl", "rFootControl"			
		};
		protected string[] keyBendingBack = {
			"headControl", "neckControl"
		};
		
		public Dictionary<string, float> targetStrength = new Dictionary<string, float>
		{
			["headControl"]       = -10
			, ["neckControl"]     = -10
			, ["chestControl"]    = -10
			, ["abdomenControl"]  = -30
			, ["abdomen2Control"] = -30
			, ["pelvisControl"]   = -30
			, ["hipControl"]      = -30
			, ["lThighControl"]   = 30
			, ["lKneeControl"]    = 30
			, ["lFootControl"]    = 30
			, ["rThighControl"]   = 30
			, ["rKneeControl"]    = 30
			, ["rFootControl"]    = 30
		};

		public Dictionary<string, float> closingTimeOffset1 = new Dictionary<string, float>
		{
			["breath"] = 1.2f
			, ["head"] = 0.3f
		};
		public Dictionary<string, float> closingTimeOffset2 = new Dictionary<string, float>
		{
			["breath"] = 1.2f
			, ["head"] = 1.5f
		};
		
		protected bool isBreathActive;
		protected bool isHeadActive;
		
		protected int waveState;
		protected bool isBellyActive;
		#endregion
		
		#region breath
		#endregion
		
		#region head
		public float bendBackwardThreshold = 0.2f;
		#endregion
		
		private float lastTime;
		
		public Belly(EBATCore core){
			atom = core.atom;
			this.core = core;
		}
		
		public void loadSettings(){
			try{
				targetStrength["chestControl"] = core.settings["General"]["BreathStrength"].AsFloat;
				targetStrength["headControl"]  = core.settings["General"]["BendingBackStrength"].AsFloat;
				targetStrength["neckControl"]  = core.settings["General"]["BendingBackStrength"].AsFloat;
			}catch (Exception e){
				core.loggerE(e);
			}

		}
		
		public void applySettings(){
			setControllerAndProgress();
		}
		
		public void maintenance(){

		}

		#region orgasm functions
		public void setControllerAndProgress(){
			try{
				ctrl   = new Dictionary<string, FreeControllerV3>();
				backup = new Dictionary<string, ControllerBackup>();
				progress = new Dictionary<string, float>();
				timing   = new Dictionary<string, CycleManager>();
				isActive = new Dictionary<string, bool>();

				initialQuat = new Dictionary<string, Quaternion>();
				targetQuat = new Dictionary<string, Quaternion>();
				
				string[] keys = new string[keyBreath.Length + keyOrgasm.Length + keyBendingBack.Length];
				Array.Copy(keyBreath, 0, keys, 0, keyBreath.Length);
				Array.Copy(keyOrgasm, 0, keys, keyBreath.Length, keyOrgasm.Length);
				Array.Copy(keyBendingBack, 0, keys, keyBreath.Length+keyOrgasm.Length, keyBendingBack.Length);
				
				for(int i = 0; i < keys.Length; i++){
					ctrl[keys[i]] = atom.GetStorableByID(keys[i]) as FreeControllerV3;
					backup.Add(keys[i], new ControllerBackup());
					backup[keys[i]].backUp(ctrl[keys[i]]);
					progress[keys[i]] = 0;
				}
				
				isActive["breath"] = false;
				isActive["head"] = false;
				timing.Add("breath", new CycleManager());					
				timing.Add("head", new CycleManager());
				
			}catch (Exception e){
				core.loggerE(e);
			}
		}

		public void storePositionAndRotation(){
			try{
				string[] keys = new string[keyBreath.Length + keyOrgasm.Length + keyBendingBack.Length];
				Array.Copy(keyBreath, 0, keys, 0, keyBreath.Length);
				Array.Copy(keyOrgasm, 0, keys, keyBreath.Length, keyOrgasm.Length);
				Array.Copy(keyBendingBack, 0, keys, keyBreath.Length+keyOrgasm.Length, keyBendingBack.Length);

				initialQuat = new Dictionary<string, Quaternion>();
				
				for(int i = 0; i < keys.Length; i++){
					backup[keys[i]].backUp(ctrl[keys[i]]);
					initialQuat[keys[i]] = ctrl[keys[i]].transform.rotation;
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void changePositionAndRotationState(){
			try{
				for(int i = 0; i < keyOrgasm.Length; i++){
					ctrl[keyOrgasm[i]].SetRotationStateFromString("Off");
//					ctrl[keyOrgasm[i]].SetPositionStateFromString("Off");
				}
				for(int i = 0; i < keyBreath.Length; i++){
					ctrl[keyBreath[i]].SetRotationStateFromString("On");
					ctrl[keyBreath[i]].SetPositionStateFromString("On");
				}
				for(int i = 0; i < keyBendingBack.Length; i++){
					ctrl[keyBendingBack[i]].SetRotationStateFromString("On");
					ctrl[keyBendingBack[i]].SetPositionStateFromString("On");
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}		
		
		public void restorePositionAndRotation(){
			try{
				for(int i = 0; i < keyOrgasm.Length; i++){
					ctrl[keyOrgasm[i]] = backup[keyOrgasm[i]].recovery(ctrl[keyOrgasm[i]]);
				}
				for(int i = 0; i < keyBreath.Length; i++){
					ctrl[keyBreath[i]] = backup[keyBreath[i]].recovery(ctrl[keyBreath[i]]);
				}
				for(int i = 0; i < keyBendingBack.Length; i++){
					ctrl[keyBendingBack[i]] = backup[keyBendingBack[i]].recovery(ctrl[keyBendingBack[i]]);
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void updatePosition(float time, string[] keys, string position){
			try{
				if(isActive[position]){
					return;
				}else{
					core.sound.playSound("breath", Time.time);
					core.expression.update(time);

					isActive[position] = true;
					float time1 = time + closingTimeOffset1[position] * UnityEngine.Random.Range(0.7f,1.0f) * AdjustInterval();
					float time2 = time1 + closingTimeOffset2[position] * UnityEngine.Random.Range(0.7f,1.0f) * AdjustInterval();
					timing[position].setTimer(time, time1, time2);

					float tmpInitialStrength;
					float tmpTargetStrength;
					FreeControllerV3 baseController = getBaseController(position);
					for(int i = 0; i < keys.Length; i++){
						tmpInitialStrength = ctrl[keys[i]].transform.localEulerAngles.x;
						tmpTargetStrength  = tmpInitialStrength + targetStrength[keys[i]] * UnityEngine.Random.Range(0.8f,1.0f) * AdjustStrength();
						targetQuat[keys[i]] = Quaternion.AngleAxis(tmpTargetStrength, baseController.transform.right) * baseController.transform.rotation;
					}
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}

		public FreeControllerV3 getBaseController(string position){
			FreeControllerV3 result;
			if(position == "head"){
				result = ctrl["headControl"];
			}else if(position == "breath"){
				result = ctrl["chestControl"];
			}else{
				result = null;
			}
			return result;
		}
		
		public void calcProgress(float time, string position){
			try{
				progress[position] = timing[position].getProgress(time);
				
				if(timing[position].getState(time) == 0){
					isActive[position] = false;
					if(position == "breath"){
						isBreathActive = false;
					}else if(position == "head"){
						isHeadActive = false;
					}
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void changeJoint(float time, string[] keys, string position){
			try{
				int state = timing[position].getState(time);
				for(int i = 0; i < keys.Length; i++){
					if(state != 0){
						ctrl[keys[i]].transform.rotation = Quaternion.Slerp(initialQuat[keys[i]], targetQuat[keys[i]], progress[position]);
					}
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}

		public float AdjustInterval(){
			float stageCoefficient = (float)core.getStage() * 0.1f;
			return (1 / (1 + stageCoefficient));
		}
		
		public float AdjustStrength(){
			float stageCoefficient = (float)core.getStage() * 0.2f;
			return 1 + stageCoefficient;
		}
		
		public void manage(float time){
			try{
				if(time == lastTime){
					time += 0.001f;
				}else if(time < lastTime){
					time = lastTime + 0.001f;
				}
				lastTime = time;

				if(core.getStage() < 1){
					isActive["breath"] = false;
					isActive["head"] = false;
					return;
				}
				if(isActive["breath"] == false){
					updatePosition(time, keyBreath, "breath");
					return;
				}

				calcProgress(time, "breath");
				changeJoint(time, keyBreath, "breath");
				
				if(isActive["head"]){
					calcProgress(time, "head");
					changeJoint(time, keyBendingBack, "head");
				}

			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void bendHerselfBack(float time){
			float ratio = UnityEngine.Random.Range(0.0f,1.0f);
			if(ratio < bendBackwardThreshold){
				updatePosition(time, keyBendingBack, "head");
			}
		}
						
		public void bellyBump(float strength, float time){
			try{
				for(int i = 0; i < keyOrgasm.Length; i++){
					if(keyOrgasm[i].StartsWith("l")){
						ctrl[keyOrgasm[i]].jointRotationDriveZTarget = targetStrength[keyOrgasm[i]] * strength * -1;
					}else if(keyOrgasm[i].StartsWith("r")){
						ctrl[keyOrgasm[i]].jointRotationDriveZTarget = targetStrength[keyOrgasm[i]] * strength;						
					}else{
						ctrl[keyOrgasm[i]].jointRotationDriveXTarget = targetStrength[keyOrgasm[i]] * strength;
					}
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void bellyDents(float strength, float time){
			try{
				for(int i = 0; i < keyOrgasm.Length; i++){
					if(keyOrgasm[i].StartsWith("l")){
						ctrl[keyOrgasm[i]].jointRotationDriveZTarget = 0f;
					}else if(keyOrgasm[i].StartsWith("r")){
						ctrl[keyOrgasm[i]].jointRotationDriveZTarget = 0f;
					}else{
						ctrl[keyOrgasm[i]].jointRotationDriveXTarget = 0f;
					}
				}			
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		#endregion
	}
}