#region using
using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
#endregion

namespace Cotyounoyume{
	public class EyesControl
	{
		#region common or UI Properties
		public Atom atom;
		public EBATCore core;
		public bool disableLog;
		
		public bool isEyecontrolActive = false;
		public bool eyecontrolOnGlobal = true;
		public bool eyecontrolRollupOnGlobal = true;
		public bool lookModeGlobal = true;	
		public float eyecontrolClosingTimeInterval = 2f;
		#endregion
		
		#region eyeControl
		protected Transform eyeControlOrig;
		protected FreeControllerV3 headControl;
		protected FreeControllerV3 eyeControl;
		protected float eyecontrolDegree;
		protected float baseDegree = 15f;
		protected float eyecontrolClosingTime;
		protected float eyecontrolOpeningTime;
		protected float eyecontrolRandomOffset = 0.2f;
		#endregion
		
		public EyesControl(EBATCore core){
			atom = core.atom;
			isEyecontrolActive = false;
			this.core = core;
		}
		
		public void loadSettings(){		
			try{
				lookModeGlobal = core.settings["General"]["LookMode"].AsBool;
				eyecontrolOnGlobal = core.settings["General"]["EyesControlOn"].AsBool;
				eyecontrolRollupOnGlobal = core.settings["General"]["EyesRollupOn"].AsBool;
				eyecontrolClosingTimeInterval = core.settings["General"]["EyesControlDurationTime"].AsFloat;
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void applySettings(){
			try{
				atom.GetStorableByID("Eyes").SetStringChooserParamValue("lookMode", core.settings["General"]["LookMode"]);
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void maintenance(){
			try{
				#region initialize eyeControl
				headControl = atom.GetStorableByID("headControl") as FreeControllerV3;
				eyeControl = atom.GetStorableByID("eyeTargetControl") as FreeControllerV3;			
				setInitialTransformOfEyeControl();
				#endregion
			}catch (Exception e){
				core.loggerE(e);
			}
		}

		#region eyeControl functions
		public void setInitialTransformOfEyeControl(){
			try{
				eyeControl.transform.position = headControl.transform.position;
				eyeControl.transform.rotation = headControl.transform.rotation;
				eyeControl.transform.Translate(0f, 0f, 5f);
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void update(float time){
			try{
				if(eyecontrolOnGlobal == false){
					return;
				}

				float delta = time - eyecontrolOpeningTime;
				if(delta < eyecontrolClosingTime){
					return;
				}
				float randomOffset = UnityEngine.Random.Range(0.1f, eyecontrolRandomOffset);
				eyecontrolClosingTime = eyecontrolClosingTimeInterval + randomOffset;
				isEyecontrolActive = true;
				
				eyecontrolOpeningTime = time;
				string direction = calcEyeDirection();
				string[] directions = direction.Split(':');
				string upDown = directions[0];
				string rightLeft = directions[1];
				float eyecontrolStrength = calcEyeStrength();
				if(upDown == "neutral" && rightLeft == "neutral"){
					eyeControlBaseSwitch("Player");
				}else{
					changeEyeControl(upDown, rightLeft, eyecontrolStrength);
					eyeControlBaseSwitch("Target");
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}

		public float calcEyeStrength(){
			try{
				float strength = 0;
				int stage;
				stage = core.getStage();
				if(stage == 0){
					strength = 0;
				}else if(stage == 1){
					strength = UnityEngine.Random.Range(0.1f, 0.2f);
				}else if(stage == 2){
					strength = UnityEngine.Random.Range(0.2f, 0.5f);
				}else if(stage == 3){
					strength = UnityEngine.Random.Range(0.2f, 0.5f);
				}else if(stage == 4){
					if(eyecontrolRollupOnGlobal){
						strength = UnityEngine.Random.Range(0.7f, 1.5f);
					}else{
						strength = UnityEngine.Random.Range(0.2f, 0.5f);
					}				
				}else{
					if(eyecontrolRollupOnGlobal){
						strength = UnityEngine.Random.Range(0.7f, 1.5f);
					}else{
						strength = UnityEngine.Random.Range(0.2f, 0.5f);
					}
				}
				return strength;
			}catch (Exception e){
				core.loggerE(e);
				return 0f;
			}
		}
		
		public string calcEyeDirection(){
			try{
				string direction = "neutral:neutral";
				
				float rand1 = UnityEngine.Random.Range(0f, 1.0f);
				float rand2 = UnityEngine.Random.Range(0f, 1.0f);
				int stage;
				stage = core.getStage();
				if(stage == 0){
					direction = "neutral:neutral";
				}else if(stage == 1){
					if(rand1 < 0.2f){
						direction = (rand2 <= 0.5f) ? "down:right" : "down:left";
					}else{
						direction = "neutral:neutral";
					}
				}else if(stage == 2){
					if(rand1 < 0.5f){
						direction = (rand2 <= 0.5f) ? "down:right" : "down:left";
					}else{
						direction = "neutral:neutral";
					}
				}else if(stage == 3){
					if(rand1 < 0.8f){
						direction = (rand2 <= 0.5f) ? "down:right" : "down:left";
					}else{
						direction = "neutral:neutral";					
					}
				}else if(stage == 4){
					if(rand1 < 0.9f){
						if(eyecontrolRollupOnGlobal){
							direction = "up:neutral";
						}else{
							direction = (rand2 <= 0.5f) ? "down:right" : "down:left";						
						}
					}else{
						direction = "neutral:neutral";					
					}
				}else{
					if(rand1 < 0.99f){
						if(eyecontrolRollupOnGlobal){
							direction = "up:neutral";
						}else{
							direction = (rand2 <= 0.5f) ? "down:right" : "down:left";						
						}
					}else{
						direction = "neutral:neutral";					
					}
				}				
				return direction;
			}catch (Exception e){
				core.loggerE(e);
				return "neutral:neutral";
			}
		}
		
		public void eyeControlBaseSwitch(string type){
			try{
				atom.GetStorableByID("Eyes").SetStringChooserParamValue("lookMode", type);
				if(type == "Player"){
					setInitialTransformOfEyeControl();
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void manage(float time){
			try{
				float delta = time - eyecontrolOpeningTime;
				if(isEyecontrolActive){
					if(delta > eyecontrolClosingTime && core.isDuringOrgasm == false){
						isEyecontrolActive = false;
						eyeControlBaseSwitch("Player");
					}
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}

		public void changeEyeControl(string upDown, string rightLeft, float strength){
			try{
				setInitialTransformOfEyeControl();
				float degreeUpDown = 0f;
				float degreeRightLeft = 0f;
				if(upDown != "neutral"){
					degreeUpDown = (upDown == "up") ? -1 * baseDegree : baseDegree/2;
					degreeUpDown *= strength;
					eyeControl.transform.RotateAround(
						headControl.transform.position, 
						headControl.transform.right, degreeUpDown); 
				}
				if(rightLeft != "neutral"){
					degreeRightLeft = (rightLeft == "left") ? -1 * baseDegree : baseDegree;	
					degreeRightLeft *= strength;
					eyeControl.transform.RotateAround(
						headControl.transform.position, 
						headControl.transform.up, degreeRightLeft); 
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		#endregion
	}
}