#region using
using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
#endregion

namespace Cotyounoyume{
	public class ClothesManager
	{
		public ClothesStorage[] clothesStorages;
		public Atom atom;
		public DAZCharacterSelector geometryForClothes;
		
		public ClothesManager(){
			clothesStorages = new ClothesStorage[6];
			for(int i = 0; i < 6; i++){
				clothesStorages[i] = new ClothesStorage();
			}
		}
		
		public void initialize(Atom atom){
			this.atom = atom;
			JSONStorable geometryStorable = this.atom.GetStorableByID("geometry");
			geometryForClothes = geometryStorable as DAZCharacterSelector;
		}
		
		public void equipClothes(int stage){
			try{
//				debug("equipClothes: stage=" + stage);
				if(stage < 0 || stage > 5){
					return;
				}
				for(int i = 0; i < clothesStorages[stage].clothes.Count; i++){
//					debug(clothesStorages[stage].clothes[i].uid);
					geometryForClothes.SetActiveClothingItem(clothesStorages[stage].clothes[i], active: true);
				}
			}catch(Exception e){
				loggerE(e);
			}
		}
		
		public void unequipClothes(int stage){
			try{
//				debug("unequipClothes: stage=" + stage);
				if(stage < 0 || stage > 5){
					return;
				}
				for(int i = 0; i < clothesStorages[stage].clothes.Count; i++){
//					debug(clothesStorages[stage].clothes[i].uid);
					geometryForClothes.SetActiveClothingItem(clothesStorages[stage].clothes[i], active: false);
				}				
			}catch(Exception e){
				loggerE(e);
			}
		}
		
		public List<DAZClothingItem> getClothes(){
			List<DAZClothingItem> clothes = geometryForClothes.clothingItems.ToList();
			List<DAZClothingItem> myClothes = new List<DAZClothingItem>();
			foreach (DAZClothingItem dAZClothingItem in clothes){
				if(dAZClothingItem.active){
					myClothes.Add(dAZClothingItem);
				}
			}
			
			return myClothes;
		}
		
		public List<string> getClothUids(int stage){
			List<string> myClothes = clothes2uids(clothesStorages[stage].clothes);			
			return myClothes;
		}
		
		public void setClothes(int stage){
			try{
				List<DAZClothingItem> tmpClohtes = getClothes();
				clothesStorages[stage].setClothes(tmpClohtes);				
			}catch(Exception e){
				loggerE(e);
			}
		}
		
		public void showClothes(int stage){
			try{
//				debug("stage:" + stage);
				clothesStorages[stage].showClothes();
			}catch(Exception e){
				loggerE(e);
			}
		}
		
		public void setClothesByUids(int stage, List<string> uids){
			try{
				List<DAZClothingItem> tmpClohtes = this.uids2clothes(uids);
				clothesStorages[stage].setClothes(tmpClohtes);
			}catch(Exception e){
				loggerE(e);
			}
		}
		
		public void clearClothes(int stage){
			clothesStorages[stage].clearClothes();
		}
		
		public void undressAll(){
			geometryForClothes.RemoveAllClothing();
		}
		
		public List<string> clothes2uids(List<DAZClothingItem> clothes){
			int num = clothes.Count;
			List<string> uids = new List<string>();
			for(int i = 0; i < num; i++){
				uids.Add(clothes[i].uid);
			}
			return uids;
		}
		
		public List<DAZClothingItem> uids2clothes(List<string> uids){
			Dictionary<string, bool> uidsList;
			uidsList = new Dictionary<string, bool>();			
			for(int i = 0; i < uids.Count; i++){
				uidsList[uids[i]] = true;
			}
			
			List<DAZClothingItem> clothes = new List<DAZClothingItem>();

			List<DAZClothingItem> clothesAll = geometryForClothes.clothingItems.ToList();
			for(int i = 0; i < clothesAll.Count; i++){
				if(uidsList.ContainsKey(clothesAll[i].uid)){
					clothes.Add(clothesAll[i]);
				}
			}
			
			return clothes;
		}
		
		public void debug(string str){
			SuperController.LogMessage("debug: " + str);
		}
		
		public void loggerE(Exception e){
			SuperController.LogMessage("*** Exception:" + e);
		}
	}
	
	
	public class ClothesStorage
	{
		public List<DAZClothingItem> clothes;
		
		public ClothesStorage(){
			clothes = new List<DAZClothingItem>();
		}
		
		public void setClothes(List<DAZClothingItem> clothes){
			this.clothes = new List<DAZClothingItem>();
			for(int i = 0; i < clothes.Count; i++){
				this.clothes.Add(clothes[i]);
			}
		}
		
		public void showClothes(){
			if(this.clothes.Count == 0){
				SuperController.LogMessage("  clothes not found.");				
			}
			for(int i = 0; i < this.clothes.Count; i++){
				SuperController.LogMessage("  Name: " + this.clothes[i].displayName);
//				SuperController.LogMessage("  Uid : " + this.clothes[i].uid);
			}
		}
		
		public void clearClothes(){
			this.clothes = new List<DAZClothingItem>();
		}
		
		public void debug(string str){
			SuperController.LogMessage("debug: " + str);
		}
		
		public void loggerE(Exception e){
			SuperController.LogMessage("*** Exception:" + e);
		}
	}
	
	public class Tears
	{
		#region common or UI Properties
		public Atom atom;
		public float tearsClosingTimeOffset = 10f;
		public EBATCore core;
		public bool disableLog;
		public bool isTearsGloss = true;
		public bool tearsOnGlobal = true;
		public bool droolOnGlobal = true;
		public bool isTearsActive = false;
		public ClothesManager cm;
		#endregion
		
		#region tears parameters
		protected float tearsClosingTime;
		protected DAZCharacterSelector geometryForClothes;
		protected Dictionary<string, DAZClothingItem> tearsMap;
		protected bool isTearsUpdated;
		protected string tearsString;
		protected string droolString;
		protected bool droolOn = false;
		protected bool tearsOn = false;	
		public List<string> tearsTypes;
		public string tearsTypeDefault = "Clear";
		public string currentTearsType = "Clear";
		public bool isTearsAlt;
		#endregion
		
		public Tears(EBATCore core){
			try{
				atom = core.atom;
				isTearsActive = false;
				this.core = core;
				cm = new ClothesManager();
				cm.initialize(atom);
			}catch(Exception e){
				core.loggerE(e);
			}
		}
		
		public void loadSettings(){		
			try{
				tearsTypeDefault = core.settings["General"]["DefaultTearsType"];
				setTearsType(tearsTypeDefault);
				tearsOnGlobal = core.settings["General"]["TearsOn"].AsBool;
				droolOnGlobal = core.settings["General"]["DroolOn"].AsBool;
				isTearsGloss = core.settings["General"]["TearsGlossOn"].AsBool;			
				tearsClosingTimeOffset = core.settings["General"]["TearsDurationTime"].AsFloat;

				JSONNode jsThreshold = core.settings["StageThreshold"];
				for(int i = 0; i < jsThreshold.Count; i++){
					int stage = i+1;
					if(jsThreshold[i]["Clothes"].Count > 0){
						List<string> uids = new List<string>();
						for(int j = 0; j < jsThreshold[i]["Clothes"].Count; j++){
							uids.Add(jsThreshold[i]["Clothes"][j]);
						}						
						cm.setClothesByUids(stage, uids);
					}
				}
				
				#region read tears from json
				tearsTypes = new List<string>();
				JSONNode jsTears = core.settings["Tears"];
				for(int i = 0; i < jsTears.Count; i++){
					tearsTypes.Add(jsTears[i]);
				}
				#endregion

			}catch(Exception e){
				core.loggerE(e);
			}
		}
		
		public void applySettings(){

		}
		
		public void setTearsType(string type){
			if(type.EndsWith("(Alt)")){
				isTearsAlt = true;
				currentTearsType = type.Replace("(ALT)", "");
			}else{
				isTearsAlt = false;
				currentTearsType = type;
			}
		}
		
		public void maintenance(){
			#region initialize tears(clothes)
			try{
				JSONStorable geometryStorable = atom.GetStorableByID("geometry");
				geometryForClothes = geometryStorable as DAZCharacterSelector;
			
				tearsMap = new Dictionary<string, DAZClothingItem>();			
				List<DAZClothingItem> clothes = geometryForClothes.clothingItems.ToList();
				for (int i = 0; i < clothes.Count; i++){
					if(clothes[i].creatorName == "cotyounoyume"){
						if(tearsMap.ContainsKey(clothes[i].displayName)){
							core.logger("debug: Tears: clothes[i].displayName:" + clothes[i].displayName + " exists.");
						}else{
							tearsMap.Add(clothes[i].displayName, clothes[i]);
						}
					}
				}
				#endregion
			}catch(Exception e){
				core.loggerE(e);
			}
		}

		#region tears functions
		public void update(float time){
			try{
				if(tearsOnGlobal == false){
					return;
				}
//				applyBodyMorph();
					
				tearsClosingTime = time + tearsClosingTimeOffset;

				string lastTearsString = tearsString;
				string lastDroolString = droolString;
				isTearsUpdated = false;
				tearsString = "Tears";
				droolString = "Drool";
				string length = "";
				string color = " " + currentTearsType;
				string type = (isTearsAlt) ? " (type2)" : " (type1)";
				
				droolOn = false;
				tearsOn = false;
				
				int stage = core.getStage();
				if(stage == 0){
					tearsOn = false;
				}else if(stage == 1){
					tearsOn = true;
					length = " Small";
				}else if(stage == 2){
					tearsOn = true;
					length = " Middle";
				}else if(stage == 3){
					tearsOn = true;
					length = " Large";
				}else if(stage >= 4){
					length = " Large";
					tearsOn = true;
					droolOn = droolOnGlobal ? true : false;
				}

				if(tearsOn){
					tearsString += length + color + type;
				}
				if(droolOn){
					droolString += " Normal" + color + " (type1)";
				}
				
				core.logger("tearsString:" + tearsString);
				core.logger("droolString:" + droolString);
				core.logger("lastTearsString:" + lastTearsString);
				core.logger("lastDroolString:" + lastDroolString);
				
				if(tearsOn){
					if(!tearsMap.ContainsKey(tearsString)){
						core.logger("tearsMap: key=" + tearsString + "not found.");
					}else{
						geometryForClothes.SetActiveClothingItem(tearsMap[tearsString], true);
					}
					isTearsActive = true;
				}
				if(droolOn){
					if(!tearsMap.ContainsKey(droolString)){
						core.logger("tearsMap: key=" + droolString + "not found.");
					}else{
						geometryForClothes.SetActiveClothingItem(tearsMap[droolString], true);								
					}
				}
				if(tearsString != lastTearsString && lastTearsString != null){
					if(!tearsMap.ContainsKey(lastTearsString)){
						core.logger("tearsMap: key=" + lastTearsString + "not found.");
					}else{
						geometryForClothes.SetActiveClothingItem(tearsMap[lastTearsString], active: false);
					}
					isTearsUpdated = true;
				}
				if(droolString != lastDroolString && lastDroolString != null){
					if(!tearsMap.ContainsKey(lastDroolString)){
						core.logger("tearsMap: key=" + lastDroolString + "not found.");
					}else{
						geometryForClothes.SetActiveClothingItem(tearsMap[lastDroolString], active: false);
					}
					isTearsUpdated = true;
				}
			}catch(Exception e){
				core.loggerE(e);
			}
		}
		
		public void unequipTears(){
			try{
				if(tearsOn){
					geometryForClothes.SetActiveClothingItem(tearsMap[tearsString], active: false);
				}
				if(droolOn){
					geometryForClothes.SetActiveClothingItem(tearsMap[droolString], active: false);				
				}
			}catch(Exception e){
				core.loggerE(e);
			}
		}
		
		public void unequipTearsAll(){
			try{
				foreach(KeyValuePair<string, DAZClothingItem> cloth in tearsMap){
					geometryForClothes.SetActiveClothingItem(cloth.Value, active: false);					
				}
			}catch(Exception e){
				core.loggerE(e);
			}
		}
		
		public void manage(float time){
			try{
				if(isTearsActive && tearsClosingTime < time && core.isDuringOrgasm == false){
					core.logger("tears effect stop.");
					isTearsActive = false;
					unequipTearsAll();
					//geometryForClothes.SetActiveClothingItem(tearsMap[tearsString], active: false);
					//geometryForClothes.SetActiveClothingItem(tearsMap[droolString], active: false);
				}
			}catch(Exception e){
				core.loggerE(e);
			}
		}
		
		public void refreshTears(){
			try{
//				unequipTears();
				unequipTearsAll();
				tearsString = "";
				droolString = "";
				update(Time.time);
			}catch(Exception e){
				core.loggerE(e);
			}
		}
		#endregion
	}
}